#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-02-21 08:58:35 (ywatanabe)"

import torch


def psd_torch(signal, samp_rate):
    """
    # Example usage:
    samp_rate = 480  # Sampling rate in Hz
    signal = torch.randn(480)  # Example signal with 480 samples
    freqs, psd = calculate_psd(signal, samp_rate)

    # Plot the PSD (if you have matplotlib installed)
    import matplotlib.pyplot as plt

    plt.plot(freqs.numpy(), psd.numpy())
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Power/Frequency (V^2/Hz)')
    plt.title('Power Spectral Density')
    plt.show()
    """
    # Apply window function to the signal (e.g., Hanning window)
    window = torch.hann_window(signal.size(-1))
    signal = signal * window

    # Perform the FFT
    fft_output = torch.fft.fft(signal)

    # Compute the power spectrum (magnitude squared of the FFT output)
    power_spectrum = torch.abs(fft_output) ** 2

    # Normalize the power spectrum to get the PSD
    # Usually, we divide by the length of the signal and the sum of the window squared
    # to get the power in terms of physical units (e.g., V^2/Hz)
    psd = power_spectrum / (samp_rate * (window**2).sum())

    # Since the signal is real, we only need the positive half of the FFT output
    # The factor of 2 accounts for the energy in the negative frequencies that we're discarding
    psd = psd[: len(psd) // 2] * 2

    # Adjust the DC component (0 Hz) and Nyquist component (if applicable)
    psd[0] /= 2
    if len(psd) % 2 == 0:  # Even length, Nyquist freq component is included
        psd[-1] /= 2

    # Frequency axis
    freqs = torch.fft.fftfreq(signal.size(-1), 1 / samp_rate)[: len(psd)]

    return freqs, psd



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-07 23:44:26 (ywatanabe)"


def hide_spines(ax, tgts=["top", "right", "bottom", "left"]):
    """
    Hides specified spines from a matplotlib Axes.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to modify.
    - tgts (list of str): List of spines to hide from the Axes.

    Returns:
    - ax (matplotlib.axes.Axes): The modified Axes object with spines removed.
    """
    for tgt in tgts:
        ax.spines[tgt].set_visible(False)

    return ax



#!/usr/bin/env python3

from .load import (
    get_data_path_from_a_package,
    load,
    load_configs,
    load_study_rdb,
    load_yaml_as_an_optuna_dict,
)
from .path import (
    find,
    find_latest,
    find_the_git_root_dir,
    get_this_fpath,
    increment_version,
    mk_spath,
    split_fpath,
    touch,
)
from .save import (
    is_listed_X,
    save,
    save_listed_dfs_as_csv,
    save_listed_scalars_as_csv,
    save_optuna_study_as_csv_and_pngs,
)



#!/usr/bin/env python3

# from . import colors
from . import ax
from ._add_hue import add_hue
from ._annotated_heatmap import annotated_heatmap
from ._configure_mpl import configure_mpl
from ._draw_a_cube import draw_a_cube
from ._get_RGBA_from_colormap import get_RGBA_from_colormap
from ._mk_colorbar import mk_colorbar
from ._mk_patches import mk_patches
from ._subplots import subplots



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-02-04 13:05:40 (ywatanabe)"

from collections import OrderedDict

import matplotlib.pyplot as plt
import mngs
import numpy as np
import pandas as pd


def set_n_ticks(axis, n_xticks=4, n_yticks=4):
    axis.xaxis.set_major_locator(plt.MaxNLocator(n_xticks))
    axis.yaxis.set_major_locator(plt.MaxNLocator(n_yticks))
    return axis


class SubplotsManager:
    """
    A manager class monitors data plotted using the ax methods from matplotlib.pyplot.
    This data can be converted into a CSV file formatted for SigmaPlot compatibility.
    Additionally, refer to the SigmaPlot macros available at mngs.gists (https://github.com/ywatanabe1989/mngs/tree/main/src/mngs/gists).

    Currently, available methods are as follows:
        ax.plot, ax.scatter, ax.boxplot, and ax.bar

    Example:
        import matplotlib
        import mngs

        matplotlib.use("Agg")  # "TkAgg"

        fig, ax = subplots()
        ax.plot([1, 2, 3], [4, 5, 6], id="plot1")
        ax.plot([4, 5, 6], [1, 2, 3], id="plot2")
        mngs.io.save(fig, "/tmp/subplots_demo/plots.png")

        # Behaves like native matplotlib.pyplot.subplots without tracking
        fig, ax = subplots(track=False)
        ax.plot([1, 2, 3], [4, 5, 6], id="plot3")
        ax.plot([4, 5, 6], [1, 2, 3], id="plot4")
        mngs.io.save(fig, "/tmp/subplots_demo/plots.png")

        fig, ax = subplots()
        ax.scatter([1, 2, 3], [4, 5, 6], id="scatter1")
        ax.scatter([4, 5, 6], [1, 2, 3], id="scatter2")
        mngs.io.save(fig, "/tmp/subplots_demo/scatters.png")

        fig, ax = subplots()
        ax.boxplot([1, 2, 3], id="boxplot1")
        mngs.io.save(fig, "/tmp/subplots_demo/boxplot1.png")

        fig, ax = subplots()
        ax.bar(["A", "B", "C"], [4, 5, 6], id="bar1")
        mngs.io.save(fig, "/tmp/subplots_demo/bar1.png")

        print(ax.to_sigma())
        #    plot1_plot_x  plot1_plot_y  plot2_plot_x  ...  boxplot1_boxplot_x  bar1_bar_x  bar1_bar_y
        # 0           1.0           4.0           4.0  ...                 1.0           A         4.0
        # 1           2.0           5.0           5.0  ...                 2.0           B         5.0
        # 2           3.0           6.0           6.0  ...                 3.0           C         6.0

        print(ax.to_sigma().keys())  # plot3 and plot 4 are not tracked
        # [3 rows x 11 columns]
        # Index(['plot1_plot_x', 'plot1_plot_y', 'plot2_plot_x', 'plot2_plot_y',
        #        'scatter1_scatter_x', 'scatter1_scatter_y', 'scatter2_scatter_x',
        #        'scatter2_scatter_y', 'boxplot1_boxplot_x', 'bar1_bar_x', 'bar1_bar_y'],
        #       dtype='object')

        # If a path is passed, the SigmaPlot-friendly dataframe is saved as a csv file.
        ax.to_sigma("./tmp/subplots_demo/for_sigmaplot.csv")
        # Saved to: ./tmp/subplots_demo/for_sigmaplot.csv
    """

    def __init__(
        self,
    ):
        """Initialize the SubplotsManager with an empty plot history."""
        # self._plot_history = {}

        self._plot_history = OrderedDict()

    def __call__(self, *args, track=True, **kwargs):
        """
        Create subplots and wrap the axes with AxisDataCollector.

        Returns:
            tuple: A tuple containing the figure and wrapped axes
            in the same manner with matplotlib.pyplot.subplots.
        """
        fig, ax = plt.subplots(*args, **kwargs)
        ax = np.atleast_1d(ax)
        ax_wrapped = [
            AxisDataCollector(a, self._plot_history, track) for a in ax
        ]
        return fig, ax_wrapped if len(ax_wrapped) > 1 else ax_wrapped[0]

    @property
    def history(self):
        """
        Get the sorted plot history.

        Returns:
            dict: The sorted plot history.
        """
        return {k: self._plot_history[k] for k in self._plot_history}

    def reset_history(self):
        """Reset the plot history to an empty state."""
        self._plot_history = OrderedDict()

    @property
    def to_sigma(self):
        """
        Convert the plot history data to a SigmaPlot-compatible format DataFrame.

        Returns:
            DataFrame: The plot history in SigmaPlot format.
        """
        data_frames = [to_sigmaplot_format(v) for v in self.history.values()]
        combined_data = pd.concat(data_frames)
        return combined_data.apply(
            lambda col: col.dropna().reset_index(drop=True)
        )


class AxisDataCollector:
    """
    A wrapper class for a Matplotlib axis that collects plotting data.

    Attributes:
        axis (matplotlib.axes.Axes): The Matplotlib axis being wrapped.
        _history (dict): A reference to the history dictionary from SubplotsManager.
    """

    def __init__(self, axis, history, track):
        """
        Initialize the AxisDataCollector with a given axis and history reference.

        Arguments:
            axis (matplotlib.axes.Axes): The Matplotlib axis to wrap.
            history (dict): A reference to the SubplotsManager's history dictionary.
        """
        self.axis = axis
        self._history = history
        self.track = track

    def __getattr__(self, attr):
        """
        Wrap the axis attribute access to collect plot calls.

        Returns:
            function: A wrapper function that calls the axis method and stores the history.
        """

        def wrapper(
            *args, id=None, track=True, n_xticks=4, n_yticks=4, **kwargs
        ):
            method = getattr(self.axis, attr)
            result = method(*args, **kwargs)

            # Apply set_n_ticks after the plotting method is called
            if method in ["plot", "scatter"]:
                self.axis = set_n_ticks(
                    self.axis,
                    n_xticks=n_xticks,
                    n_yticks=n_yticks,
                )

            # Only store the history if tracking is enabled and an ID is provided
            if self.track and (id is not None):
                self._history[id] = (id, attr, args, kwargs)
            return result

        return wrapper

    @property
    def history(self):
        """
        Get the sorted history for this axis.

        Returns:
            dict: The sorted history for this axis.
        """
        return {k: self._history[k] for k in self._history}

    def reset_history(self):
        """Reset the history for this axis."""
        self._history = {}

    def to_sigma(self, lpath=None):
        """
        Convert the axis history to a sigma format DataFrame.

        Returns:
            DataFrame: The axis history in sigma format.
        """
        try:
            data_frames = [
                to_sigmaplot_format(v) for v in self.history.values()
            ]
            combined_data = pd.concat(data_frames)
            combined_data = combined_data.apply(
                lambda col: col.dropna().reset_index(drop=True)
            )

        except Exception as e:
            print(e)
            combined_data = pd.DataFrame()

        if lpath is not None:
            mngs.io.save(combined_data, lpath)

        return combined_data


def to_sigmaplot_format(record):
    """
    Convert a single plot record to a sigma format DataFrame.

    Arguments:
        record (tuple): A tuple containing the plot id, method, arguments, and keyword arguments.

    Returns:
        DataFrame: The plot data in sigma format.
    """

    id, method, args, kwargs = record
    if method in ["plot", "scatter"]:
        x, y = args
        df = pd.DataFrame({f"{id}_{method}_x": x, f"{id}_{method}_y": y})
    elif method == "bar":
        x, y = args
        df = pd.DataFrame({f"{id}_{method}_x": x, f"{id}_{method}_y": y})
    elif method == "boxplot":
        x = args[0]
        df = pd.DataFrame({f"{id}_{method}_x": x})
    return df


subplots = SubplotsManager()


if __name__ == "__main__":
    import matplotlib
    import mngs

    matplotlib.use("Agg")  # "TkAgg"

    fig, ax = subplots()
    ax.plot([1, 2, 3], [4, 5, 6], id="plot1")
    ax.plot([4, 5, 6], [1, 2, 3], id="plot2")
    mngs.io.save(fig, "/tmp/subplots_demo/plots.png")

    # Behaves like native matplotlib.pyplot.subplots without tracking
    fig, ax = subplots(track=False)
    ax.plot([1, 2, 3], [4, 5, 6], id="plot3")
    ax.plot([4, 5, 6], [1, 2, 3], id="plot4")
    mngs.io.save(fig, "/tmp/subplots_demo/plots.png")

    fig, ax = subplots()
    ax.scatter([1, 2, 3], [4, 5, 6], id="scatter1")
    ax.scatter([4, 5, 6], [1, 2, 3], id="scatter2")
    mngs.io.save(fig, "/tmp/subplots_demo/scatters.png")

    fig, ax = subplots()
    ax.boxplot([1, 2, 3], id="boxplot1")
    mngs.io.save(fig, "/tmp/subplots_demo/boxplot1.png")

    fig, ax = subplots()
    ax.bar(["A", "B", "C"], [4, 5, 6], id="bar1")
    mngs.io.save(fig, "/tmp/subplots_demo/bar1.png")

    print(ax.to_sigma())
    #    plot1_plot_x  plot1_plot_y  plot2_plot_x  ...  boxplot1_boxplot_x  bar1_bar_x  bar1_bar_y
    # 0           1.0           4.0           4.0  ...                 1.0           A         4.0
    # 1           2.0           5.0           5.0  ...                 2.0           B         5.0
    # 2           3.0           6.0           6.0  ...                 3.0           C         6.0

    print(ax.to_sigma().keys())  # plot3 and plot 4 are not tracked
    # [3 rows x 11 columns]
    # Index(['plot1_plot_x', 'plot1_plot_y', 'plot2_plot_x', 'plot2_plot_y',
    #        'scatter1_scatter_x', 'scatter1_scatter_y', 'scatter2_scatter_x',
    #        'scatter2_scatter_y', 'boxplot1_boxplot_x', 'bar1_bar_x', 'bar1_bar_y'],
    #       dtype='object')

    # If a path is passed, the sigmaplot-friendly dataframe is saved as a csv file.
    ax.to_sigma("./tmp/subplots_demo/for_sigmaplot.csv")
    # Saved to: ./tmp/subplots_demo/for_sigmaplot.csv



#!/usr/bin/env python3

import json
import logging
import os
import pickle
import warnings
from glob import glob

import h5py
import joblib  # [REVISED]
import mne
import mngs
import numpy as np
import pandas as pd
import torch
import yaml

if "general" in __file__:
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        warnings.warn(
            '\n"mngs.general.load" will be removed. '
            'Please use "mngs.io.load" instead.',
            PendingDeprecationWarning,
        )


def load(lpath, show=False, **kwargs):
    try:
        extension = "." + lpath.split(".")[-1]  # [REVISED]

        # csv
        if extension == ".csv":
            obj = pd.read_csv(lpath, **kwargs)
            obj = obj.loc[
                :, ~obj.columns.str.contains("^Unnamed")
            ]  # [REVISED]
        # tsv
        elif extension == ".tsv":
            obj = pd.read_csv(lpath, sep="\t", **kwargs)

        # excel
        elif extension in [".xls", ".xlsx", ".xlsm", ".xlsb"]:  # [REVISED]
            obj = pd.read_excel(lpath, **kwargs)

        # numpy
        elif extension == ".npy":
            obj = np.load(lpath, allow_pickle=True, **kwargs)  # [REVISED]
        # pkl
        elif extension == ".pkl":
            with open(lpath, "rb") as l:
                obj = pickle.load(l, **kwargs)
        # joblib
        elif extension == ".joblib":
            with open(lpath, "rb") as l:
                obj = joblib.load(l, **kwargs)
        # hdf5
        elif extension == ".hdf5":
            obj = {}
            with h5py.File(lpath, "r") as hf:
                for name in hf:  # [REVISED]
                    obj[name] = hf[name][:]
        # json
        elif extension == ".json":
            with open(lpath, "r") as f:
                obj = json.load(f)
        # png
        elif extension == ".png":
            pass
        # tiff
        elif extension in [".tiff", ".tif"]:
            pass
        # yaml
        elif extension == ".yaml":
            # from ruamel.yaml import YAML

            # yaml = YAML()
            # yaml.preserve_quotes = (
            #     True  # Optional: if you want to preserve quotes
            # )
            # yaml.indent(
            #     mapping=2, sequence=4, offset=2
            # )  # Optional: set indentation

            lower = kwargs.pop("lower", False)

            with open(lpath) as f:
                obj = yaml.safe_load(f, **kwargs)  # [REVISED]

            if lower:
                obj = {k.lower(): v for k, v in obj.items()}

        # txt
        elif extension in [".txt", ".log", ".event"]:
            with open(lpath, "r") as f:  # [REVISED]
                obj = f.read().splitlines()  # [REVISED]
        # pth
        elif extension in [".pth", ".pt"]:
            obj = torch.load(lpath, **kwargs)

        # mat
        elif extension == ".mat":  # [REVISED]
            from pymatreader import read_mat  # [REVISED]

            obj = read_mat(lpath, **kwargs)  # [REVISED]
        # xml
        elif extension == ".xml":  # [REVISED]
            from xml2dict import xml2dict  # [REVISED]

            obj = xml2dict(lpath, **kwargs)  # [REVISED]
        # # edf
        # elif extension == ".edf":  # [REVISED]
        #     obj = mne.io.read_raw_edf(lpath, preload=True)  # [REVISED]
        # con
        elif extension == ".con":  # [REVISED]
            obj = mne.io.read_raw_fif(
                lpath, preload=True, **kwargs
            )  # [REVISED]
            obj = obj.to_data_frame()  # [REVISED]
            obj["samp_rate"] = obj.info["sfreq"]  # [REVISED]
        # # mrk
        # elif extension == ".mrk":  # [REVISED]
        #     obj = mne.io.read_mrk(lpath)  # [REVISED]

        # catboost model
        elif extension == ".cbm":  # [REVISED]
            from catboost import CatBoostModel  # [REVISED]

            obj = CatBoostModel.load_model(lpath, **kwargs)  # [REVISED]

        # EEG data
        elif extension in [
            ".vhdr",
            ".vmrk",
            ".edf",
            ".bdf",
            ".gdf",
            ".cnt",
            ".egi",
            ".eeg",
            ".set",
        ]:
            obj = load_eeg_data(lpath, **kwargs)

        else:
            print(f"\nNot loaded from: {lpath}\n")
            return None

        if show:
            print(f"\nLoaded from: {lpath}\n")

        return obj

    except Exception as e:
        logging.error(f"\n{lpath} was not loaded:\n{e}")
        return None


def load_eeg_data(filename, **kwargs):
    """
    Load EEG data based on file extension and associated files using MNE-Python.

    Parameters:
    filename (str: The path to the file to be loaded.

    Returns:
    raw (mne.io.Raw: The loaded raw EEG data.
    """
    # Get the file extension
    extension = filename.split(".")[-1]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        # Load the data based on the file extension
        if extension in ["vhdr", "vmrk"]:
            # Load BrainVision data
            raw = mne.io.read_raw_brainvision(filename, preload=True, **kwargs)
        elif extension == "edf":
            # Load European data format
            raw = mne.io.read_raw_edf(filename, preload=True, **kwargs)
        elif extension == "bdf":
            # Load BioSemi data format
            raw = mne.io.read_raw_bdf(filename, preload=True, **kwargs)
        elif extension == "gdf":
            # Load General data format
            raw = mne.io.read_raw_gdf(filename, preload=True, **kwargs)
        elif extension == "cnt":
            # Load Neuroscan CNT data
            raw = mne.io.read_raw_cnt(filename, preload=True, **kwargs)
        elif extension == "egi":
            # Load EGI simple binary data
            raw = mne.io.read_raw_egi(filename, preload=True, **kwargs)
        elif extension == "set":
            # ???
            raw = mne.io.read_raw(filename, preload=True, **kwargs)
        elif extension == "eeg":
            is_BrainVision = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".vhdr", ".vmrk"]
            )
            is_NihonKoden = any(
                os.path.isfile(filename.replace(".eeg", ext))
                for ext in [".21e", ".pnt", ".log"]
            )

            # Brain Vision
            if is_BrainVision:
                filename_v = filename.replace(".eeg", ".vhdr")
                raw = mne.io.read_raw_brainvision(
                    filename_v, preload=True, **kwargs
                )
            # Nihon Koden
            if is_NihonKoden:
                # raw = mne.io.read_raw_nihon(filename, preload=True, **kwargs)
                raw = mne.io.read_raw(filename, preload=True, **kwargs)
        else:
            raise ValueError(f"Unsupported file extension: {extension}")

        return raw


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def get_data_path_from_a_package(package_str, resource):
    import importlib
    import os
    import sys

    spec = importlib.util.find_spec(package_str)
    data_dir = os.path.join(spec.origin.split("src")[0], "data")
    resource_path = os.path.join(data_dir, resource)
    return resource_path


def load_yaml_as_an_optuna_dict(fpath_yaml, trial):
    _d = load(fpath_yaml)

    for k, v in _d.items():

        dist = v["distribution"]

        if dist == "categorical":
            _d[k] = trial.suggest_categorical(k, v["values"])

        elif dist == "uniform":
            _d[k] = trial.suggest_int(k, float(v["min"]), float(v["max"]))

        elif dist == "loguniform":
            _d[k] = trial.suggest_loguniform(
                k, float(v["min"]), float(v["max"])
            )

        elif dist == "intloguniform":
            _d[k] = trial.suggest_int(
                k, float(v["min"]), float(v["max"]), log=True
            )

    return _d


def load_study_rdb(study_name, rdb_raw_bytes_url):
    """
    study = load_study_rdb(
        study_name="YOUR_STUDY_NAME",
        rdb_raw_bytes_url="sqlite:///*.db"
    )
    """
    import optuna

    # rdb_raw_bytes_url = "sqlite:////tmp/fake/ywatanabe/_MicroNN_WindowSize-1.0-sec_MaxEpochs_100_2021-1216-1844/optuna_study_test_file#0.db"
    storage = optuna.storages.RDBStorage(url=rdb_raw_bytes_url)
    study = optuna.load_study(study_name=study_name, storage=storage)
    print(f"\nLoaded: {rdb_raw_bytes_url}\n")
    return study


def load_configs(IS_DEBUG=None, show=False):
    if os.getenv("CI") == "true":
        IS_DEBUG = True

    def update_debug(config, IS_DEBUG):
        if IS_DEBUG:
            debug_keys = mngs.gen.search("^DEBUG_", list(config.keys()))[1]
            for dk in debug_keys:
                dk_wo_debug_prefix = dk.split("DEBUG_")[1]
                config[dk_wo_debug_prefix] = config[dk]
                if show:
                    print(f"\n{dk} -> {dk_wo_debug_prefix}\n")
        return config

    # Check ./config/IS_DEBUG.yaml file if IS_DEBUG argument is not passed
    if IS_DEBUG is None:
        try:
            IS_DEBUG = mngs.io.load("./config/IS_DEBUG.yaml")["IS_DEBUG"]
        except Exception as e:
            print(e)
            IS_DEBUG = False

    # Main
    CONFIGS = {}
    for lpath in glob("./config/*.yaml"):
        CONFIG = update_debug(mngs.io.load(lpath), IS_DEBUG)
        CONFIGS.update(CONFIG)

    return CONFIGS



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-07 23:38:32 (ywatanabe)"

from ._circular_hist import circular_hist
from ._extend import extend
from ._fill_between import fill_between
from ._hide_spines import hide_spines
from ._map_ticks import map_ticks
from ._panel import panel
from ._sci_note import sci_note
from ._set_n_ticks import set_n_ticks
from ._set_pos import set_pos
from ._set_size import set_size



#!/usr/bin/env python3


def set_pos(
    fig, ax, x_delta_offset_cm, y_delta_offset_cm, dragh=False, dragv=False
):
    """
    Adjusts the position of an Axes object within a Figure by a specified offset in centimeters.

    Parameters:
    - fig (matplotlib.figure.Figure): The Figure object containing the Axes.
    - ax (matplotlib.axes.Axes): The Axes object to modify.
    - x_delta_offset_cm (float): The horizontal offset in centimeters to adjust the Axes position.
    - y_delta_offset_cm (float): The vertical offset in centimeters to adjust the Axes position.
    - dragh (bool): If True, reduces the width of the Axes by the horizontal offset.
    - dragv (bool): If True, reduces the height of the Axes by the vertical offset.

    Returns:
    - ax (matplotlib.axes.Axes): The modified Axes object with the adjusted position.
    """

    bbox = ax.get_position()

    # Calculates delta ratios
    fig_width_inch, fig_height_inch = fig.get_size_inches()

    x_delta_offset_inch = x_delta_offset_cm / 2.54
    y_delta_offset_inch = y_delta_offset_cm / 2.54

    x_delta_offset_ratio = x_delta_offset_inch / fig_width_inch
    y_delta_offset_ratio = y_delta_offset_inch / fig_height_inch  # [REVISED]

    # Determines updated bbox position
    left = bbox.x0 + x_delta_offset_ratio
    bottom = bbox.y0 + y_delta_offset_ratio
    width = bbox.width  # [REVISED]
    height = bbox.height  # [REVISED]

    if dragh:
        width -= x_delta_offset_ratio

    if dragv:
        height -= y_delta_offset_ratio

    ax.set_position([left, bottom, width, height])  # [REVISED]

    return ax


# def set_pos(
#     fig,
#     ax,
#     x_delta_offset_cm,
#     y_delta_offset_cm,
#     dragh=False,
#     dragv=False,
# ):

#     bbox = ax.get_position()

#     ## Calculates delta ratios
#     fig_width_inch, fig_height_inch = fig.get_size_inches()

#     x_delta_offset_inch = float(x_delta_offset_cm) / 2.54
#     y_delta_offset_inch = float(y_delta_offset_cm) / 2.54

#     x_delta_offset_ratio = x_delta_offset_inch / fig_width_inch
#     y_delta_offset_ratio = y_delta_offset_inch / fig_width_inch

#     ## Determines updated bbox position
#     left = bbox.x0 + x_delta_offset_ratio
#     bottom = bbox.y0 + y_delta_offset_ratio
#     width = bbox.x1 - bbox.x0
#     height = bbox.y1 - bbox.y0

#     if dragh:
#         width -= x_delta_offset_ratio

#     if dragv:
#         height -= y_delta_offset_ratio

#     ax.set_pos(
#         [
#             left,
#             bottom,
#             width,
#             height,
#         ]
#     )

#     return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-05 16:27:11 (ywatanabe)"

import numpy as np
from sklearn.ensemble import GradientBoostingClassifier, RandomForestClassifier
from sklearn.linear_model import RidgeClassifierCV
from sklearn.pipeline import make_pipeline
from sklearn.svm import SVC
from sktime.classification.deep_learning.cnn import CNNClassifier
from sktime.classification.deep_learning.inceptiontime import (
    InceptionTimeClassifier,
)
from sktime.classification.deep_learning.lstmfcn import LSTMFCNClassifier
from sktime.classification.dummy import DummyClassifier
from sktime.classification.feature_based import TSFreshClassifier
from sktime.classification.hybrid import HIVECOTEV2
from sktime.classification.interval_based import TimeSeriesForestClassifier
from sktime.classification.kernel_based import RocketClassifier, TimeSeriesSVC
from sktime.transformations.panel.reduce import Tabularizer
from sktime.transformations.panel.rocket import Rocket


# rocket_pipeline = make_pipeline(
#     Rocket(n_jobs=-1),
#     RidgeClassifierCV(alphas=np.logspace(-3, 3, 10)),
# )
def rocket_pipeline(*args, **kwargs):
    return make_pipeline(
        # Rocket(n_jobs=-1),
        Rocket(*args, **kwargs),
        SVC(probability=True, kernel="linear"),
    )


GB_pipeline = make_pipeline(
    Tabularizer(),
    GradientBoostingClassifier(),
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-03 02:12:25 (ywatanabe)"

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import umap.umap_ as umap_orig
from sklearn.preprocessing import LabelEncoder


def umap(
    data_all,
    labels_all,
    axes_titles=None,
    supervised=False,
    title="UMAP Clustering",
    alpha=0.1,
    s=3,
):

    assert len(data_all) == len(labels_all)

    if isinstance(data_all, list):
        data_all = list(data_all)
        labels_all = list(labels_all)

    le = LabelEncoder()

    labels_all_orig = [np.array(labels) for labels in labels_all]

    le.fit(np.hstack(labels_all))
    labels_all = [le.transform(labels) for labels in labels_all]

    umap_model = umap_orig.UMAP(random_state=42)

    if supervised:
        _umap = umap_model.fit(data_all[0], y=labels_all[0])
        title = f"(Supervised) {title}"
    else:
        _umap = umap_model.fit(data_all[0])
        title = f"(Unsupervised) {title}"

    fig, axes = plt.subplots(ncols=len(data_all) + 1, sharex=True, sharey=True)

    for ii, (data, labels) in enumerate(zip(data_all, labels_all)):
        embedding = _umap.transform(data)

        ax = axes[ii + 1]

        axes[0].set_title("Superimposed")
        axes[0].set_aspect("equal")
        palette = "viridis"

        sns.scatterplot(
            x=embedding[:, 0],
            y=embedding[:, 1],
            hue=le.inverse_transform(labels),
            ax=axes[0],
            palette=palette,
            legend="full" if ii == 0 else False,
            s=s,
            alpha=alpha,
        )

        sns.scatterplot(
            x=embedding[:, 0],
            y=embedding[:, 1],
            hue=le.inverse_transform(labels),
            ax=ax,
            palette=palette,
            s=s,
            alpha=alpha,
        )
        ax.set_aspect("equal")

        if axes_titles is not None:
            ax.set_title(axes_titles[ii])

    legend_figs = []
    for i, ax in enumerate(axes):
        legend = ax.get_legend()
        if legend:
            legend_fig = plt.figure(figsize=(3, 2))
            new_legend = legend_fig.gca().legend(
                handles=legend.legendHandles, labels=legend.texts, loc="center"
            )
            legend_fig.canvas.draw()
            legend_filename = f"legend_{i}.png"
            legend_fig.savefig(legend_filename, bbox_inches="tight")
            legend_figs.append(legend_fig)
            plt.close(legend_fig)

    for ax in axes:
        ax.legend_ = None
        # ax.remove_legend()

    fig.suptitle(title)
    fig.supxlabel("UMAP 1")
    fig.supylabel("UMAP 2")

    return fig, legend_figs, _umap


# def umap(
#     data_all,
#     labels_all,
#     axes_titles=None,
#     supervised=False,
#     title="UMAP Clustering",
#     alpha=0.1,
#     s=3,
# ):
#     """
#     Performs UMAP clustering on the given data and labels, and generates a plot with the results.

#     Parameters:
#     - data_all (list of array-like): List of datasets to be used for UMAP clustering.
#     - labels_all (list of array-like): List of label arrays corresponding to the datasets.
#     - axes_titles (list of str, optional): Titles for each subplot axis.
#     - supervised (bool, optional): Whether to use supervised dimensionality reduction. Defaults to False.
#     - title (str, optional): Title for the entire plot. Defaults to "UMAP Clustering".
#     - alpha (float, optional): Alpha value for the scatter plot points. Defaults to 0.1.
#     - s (int, optional): Size of the scatter plot points. Defaults to 3.

#     Returns:
#     - fig (matplotlib.figure.Figure): The main figure object containing the UMAP plots.
#     - legend_figs (list of matplotlib.figure.Figure): List of figures containing the legends.
#     - _umap (umap.UMAP): The fitted UMAP model.
#     """
#     import matplotlib.pyplot as plt
#     import numpy as np
#     import seaborn as sns
#     import umap.umap_ as umap_orig
#     from sklearn.preprocessing import LabelEncoder

#     assert len(data_all) == len(labels_all)

#     if isinstance(data_all, list):
#         data_all = list(data_all)
#         labels_all = list(labels_all)

#     le = LabelEncoder()

#     labels_all_orig = [np.array(labels) for labels in labels_all]
#     if labels_all is not None:
#         labels_all = [le.fit_transform(labels) for labels in labels_all]

#     umap_model = umap_orig.UMAP(random_state=42)

#     if supervised:
#         _umap = umap_model.fit(data_all[0], y=labels_all[0])
#         title = f"(Supervised) {title}"
#     else:
#         _umap = umap_model.fit(data_all[0])
#         title = f"(Unsupervised) {title}"

#     fig, axes = plt.subplots(ncols=len(data_all) + 1, sharex=True, sharey=True)

#     for ii, (data, labels) in enumerate(zip(data_all, labels_all)):
#         embedding = _umap.transform(data)
#         ax = axes[ii + 1]

#         axes[0].set_title("Superimposed")
#         axes[0].set_aspect("equal")
#         palette = "viridis"

#         sns.scatterplot(
#             x=embedding[:, 0],
#             y=embedding[:, 1],
#             hue=le.inverse_transform(labels),
#             ax=axes[0],
#             palette=palette,
#             legend="full" if ii == 0 else False,
#             s=s,
#             alpha=alpha,
#         )

#         sns.scatterplot(
#             x=embedding[:, 0],
#             y=embedding[:, 1],
#             hue=le.inverse_transform(labels),
#             ax=ax,
#             palette=palette,
#             s=s,
#             alpha=alpha,
#         )
#         ax.set_aspect("equal")

#         if axes_titles is not None:
#             ax.set_title(axes_titles[ii])

#     legend_figs = []
#     for i, ax in enumerate(axes):
#         legend = ax.get_legend()
#         if legend:
#             legend_fig = plt.figure(figsize=(3, 2))
#             new_legend = legend_fig.gca().add_artist(legend)  # [REVISED]
#             legend.set_bbox_to_anchor((0, 0, 1, 1))
#             legend_fig.canvas.draw()
#             legend_filename = f"legend_{i}.png"
#             legend_fig.savefig(legend_filename, bbox_inches="tight")
#             legend_figs.append(legend_fig)
#             plt.close(legend_fig)

#     fig.suptitle(title)
#     fig.supxlabel("UMAP 1")
#     fig.supylabel("UMAP 2")

#     return fig, legend_figs, _umap


# # def umap(
# #     data_all,
# #     labels_all,
# #     axes_titles=None,
# #     supervised=False,
# #     title="UMAP Clustering",
# #     alpha=0.1,
# #     s=3,
# # ):

# #     assert len(data_all) == len(labels_all)

# #     if isinstance(data_all, list):
# #         data_all = list(data_all)
# #         labels_all = list(labels_all)

# #     le = LabelEncoder()

# #     # Store original labels
# #     labels_all_orig = [np.array(labels) for labels in labels_all]
# #     if labels_all is not None:
# #         labels_all = [le.fit_transform(labels) for labels in labels_all]

# #     umap_model = umap_orig.UMAP(random_state=42)

# #     # Process the primary dataset
# #     if supervised:
# #         _umap = umap_model.fit(data_all[0], y=labels_all[0])
# #         title = f"(Supervised) {title}"
# #     else:
# #         _umap = umap_model.fit(data_all[0])
# #         title = f"(Unsupervised) {title}"

# #     fig, axes = plt.subplots(ncols=len(data_all) + 1, sharex=True, sharey=True)

# #     for ii, (data, labels) in enumerate(zip(data_all, labels_all)):
# #         embedding = _umap.transform(data)
# #         ax = axes[ii + 1]

# #         # Superimposed
# #         axes[0].set_title("Superimposed")
# #         axes[0].set_aspect("equal")
# #         palette = "viridis"

# #         sns.scatterplot(
# #             x=embedding[:, 0],
# #             y=embedding[:, 1],
# #             hue=le.inverse_transform(labels),
# #             ax=axes[0],
# #             palette=palette,
# #             legend="full" if ii == 0 else False,
# #             s=s,
# #             alpha=alpha,
# #         )

# #         # Each data
# #         sns.scatterplot(
# #             x=embedding[:, 0],
# #             y=embedding[:, 1],
# #             hue=le.inverse_transform(labels),
# #             ax=ax,
# #             palette=palette,
# #             s=s,
# #             alpha=alpha,
# #         )
# #         ax.set_aspect("equal")

# #         if axes_titles is not None:
# #             ax.set_title(axes_titles[ii])

# #     # Save legends as separate figures and store them in a list
# #     legend_figs = []
# #     for i, ax in enumerate(axes):
# #         # Extract the legend from the current axis
# #         legend = ax.get_legend()
# #         if legend:
# #             # Create a new figure for the legend
# #             legend_fig = plt.figure(figsize=(3, 2))
# #             new_legend = legend_fig._get_axes().add_artist(legend)
# #             legend.set_bbox_to_anchor((0, 0, 1, 1))
# #             legend_fig.canvas.draw()
# #             # Save the legend as a PNG file
# #             legend_filename = f"legend_{i}.png"
# #             legend_fig.savefig(legend_filename, bbox_inches="tight")
# #             # Store the legend figure in the list
# #             legend_figs.append(legend_fig)
# #             plt.close(legend_fig)  # Close the figure to free memory

# #     fig.suptitle(title)
# #     fig.supxlabel("UMAP 1")
# #     fig.supylabel("UMAP 2")

# #     # Return the main figure, the list of legend figures, and the UMAP model
# #     return fig, legend_figs, _umap


# # # from itertools import cycle

# # # import matplotlib.pyplot as plt
# # # import mngs
# # # import numpy as np
# # # import seaborn as sns
# # # import umap.umap_ as umap_orig
# # # from sklearn.preprocessing import LabelEncoder


# # # def umap(
# # #     data_all,
# # #     labels_all,
# # #     axes_titles=None,
# # #     supervised=False,
# # #     title="UMAP Clustering",
# # #     alpha=0.1,
# # #     s=3,
# # #     # colors=None,
# # # ):

# # #     assert len(data_all) == len(labels_all)

# # #     if isinstance(data_all, list):
# # #         data_all = list(data_all)
# # #         labels_all = list(labels_all)

# # #     le = mngs.ml.utils.LabelEncoder()
# # #     # le = LabelEncoder()

# # #     # Store original labels
# # #     labels_all_orig = [np.array(labels) for labels in labels_all]
# # #     if labels_all is not None:
# # #         # labels_uq = np.unique(np.hstack(labels_all))
# # #         # le.fit(labels_uq)
# # #         labels_all = [le.fit_transform(labels) for labels in labels_all]

# # #     umap_model = umap_orig.UMAP(random_state=42)

# # #     # Process the primary dataset
# # #     if supervised:
# # #         _umap = umap_model.fit(data_all[0], y=labels_all[0])
# # #         title = f"(Supervised) {title}"
# # #     else:
# # #         _umap = umap_model.fit(data_all[0])
# # #         title = f"(Unsupervised) {title}"

# # #     fig, axes = plt.subplots(ncols=len(data_all) + 1, sharex=True, sharey=True)
# # #     # # Create a color palette that maps each unique label to a color
# # #     # unique_labels = np.unique(np.hstack(labels_all_orig))
# # #     # if colors is not None:
# # #     #     palette = dict(zip(unique_labels, cycle(colors)))
# # #     # else:
# # #     #     palette = sns.color_palette("hsv", len(unique_labels))
# # #     #     palette = dict(zip(unique_labels, palette))

# # #     # if colors is not None:
# # #     #     color_cycle = cycle(colors)
# # #     # else:
# # #     #     color_cycle = None

# # #     # for ii, (data, labels, labels_orig) in enumerate(
# # #     #     zip(data_all, labels_all, labels_all_orig)
# # #     # ):
# # #     for ii, (data, labels) in enumerate(zip(data_all, labels_all)):
# # #         embedding = _umap.transform(data)
# # #         ax = axes[ii + 1]

# # #         # Superimposed
# # #         axes[0].set_title("Superimposed")
# # #         axes[0].set_aspect("equal")
# # #         palette = "viridis"
# # #         # if color_cycle:
# # #         #     palette = sns.color_palette(
# # #         #         [next(color_cycle) for _ in range(len(np.unique(labels)))]
# # #         #     )
# # #         # else:
# # #         #     palette = "viridis"

# # #         sns.scatterplot(
# # #             x=embedding[:, 0],
# # #             y=embedding[:, 1],
# # #             hue=le.inverse_transform(labels),
# # #             ax=axes[0],
# # #             palette=palette,
# # #             legend="full" if ii == 0 else False,
# # #             s=s,
# # #             alpha=alpha,
# # #         )

# # #         # Each data
# # #         sns.scatterplot(
# # #             x=embedding[:, 0],
# # #             y=embedding[:, 1],
# # #             hue=le.inverse_transform(labels),
# # #             ax=ax,
# # #             palette=palette,
# # #             s=s,
# # #             alpha=alpha,
# # #         )
# # #         ax.set_aspect("equal")

# # #         if axes_titles is not None:
# # #             ax.set_title(axes_titles[ii])

# # #     # Remove the legends from the individual axes
# # #     for ax in axes:
# # #         ax.legend(loc="upper left")
# # #         legend = ax.get_legend()
# # #         if legend:
# # #             legend_fig = plt.figure(figsize=(3, 2))


# # #     fig.suptitle(title)
# # #     fig.supxlabel("UMAP 1")
# # #     fig.supylabel("UMAP 2")

# # #     return fig, _umap



#!/usr/bin/env python3

import csv
import warnings

import mngs
import numpy as np
import pandas as pd
import scipy

if "general" in __file__:
    with warnings.catch_warnings():
        warnings.simplefilter("always")
        warnings.warn(
            '\n"mngs.general.save" will be removed. '
            'Please use "mngs.io.save" instead.',
            PendingDeprecationWarning,
        )


def save(obj, sfname_or_spath, makedirs=True, show=True, **kwargs):
    """
    Example
      save(arr, 'data.npy')
      save(df, 'df.csv')
      save(serializable, 'serializable.pkl')
    """
    import inspect
    import json
    import os
    import pickle

    import h5py
    import numpy as np
    import pandas as pd
    import torch
    import yaml

    spath, sfname = None, None

    if "/" in sfname_or_spath:
        spath = sfname_or_spath
    else:
        sfname = sfname_or_spath

    if (spath is None) and (sfname is not None):
        ## for ipython
        __file__ = inspect.stack()[1].filename
        if "ipython" in __file__:
            __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

        ## spath
        fpath = __file__
        fdir, fname, _ = mngs.general.split_fpath(fpath)
        sdir = fdir + fname + "/"
        spath = sdir + sfname
        # spath = mk_spath(sfname, makedirs=True)

    ## Make directory
    if makedirs:
        sdir = os.path.dirname(spath)
        os.makedirs(sdir, exist_ok=True)

    ## Saves
    try:
        ## copy files
        is_copying_files = (
            isinstance(obj, str) or is_listed_X(obj, str)
        ) and (isinstance(spath, str) or is_listed_X(spath, str))
        if is_copying_files:
            mngs.general.copy_files(obj, spath)

        # csv
        elif spath.endswith(".csv"):
            if isinstance(obj, pd.DataFrame):  # DataFrame
                obj.to_csv(spath)
            if is_listed_X(obj, [int, float]):  # listed scalars
                save_listed_scalars_as_csv(
                    obj,
                    spath,
                    **kwargs,
                )
            if is_listed_X(obj, pd.DataFrame):  # listed DataFrame
                # save_listed_dfs_as_csv(obj, spath, indi_suffix=None, overwrite=False)
                save_listed_dfs_as_csv(obj, spath, **kwargs)
        # numpy
        elif spath.endswith(".npy"):
            np.save(spath, obj)
        # pkl
        elif spath.endswith(".pkl"):
            with open(spath, "wb") as s:  # 'w'
                pickle.dump(obj, s)
        # joblib
        elif spath.endswith(".joblib"):
            with open(spath, "wb") as s:  # 'w'
                joblib.dump(obj, s, compress=3)
        # png
        elif spath.endswith(".png"):
            # plotly
            if str(type(obj)) == "<class 'plotly.graph_objs._figure.Figure'>":
                obj.write_image(file=spath, format="png")
            # matplotlib
            else:
                try:
                    obj.savefig(spath)
                except:
                    obj.figure.savefig(spath)
            del obj
        # tiff
        elif spath.endswith(".tiff") or spath.endswith(".tif"):
            obj.savefig(
                spath, dpi=300, format="tiff"
            )  # obj is matplotlib.pyplot object
            del obj
        # mp4
        elif spath.endswith(".mp4"):
            mk_mp4(obj, spath)  # obj is matplotlib.pyplot.figure object
            del obj

        # yaml
        elif spath.endswith(".yaml"):
            from ruamel.yaml import YAML

            yaml = YAML()
            yaml.preserve_quotes = (
                True  # Optional: if you want to preserve quotes
            )
            yaml.indent(
                mapping=4, sequence=4, offset=4
            )  # Optional: set indentation

            with open(spath, "w") as f:
                yaml.dump(obj, f)

        # json
        elif spath.endswith(".json"):
            with open(spath, "w") as f:
                json.dump(obj, f, indent=4)

        # hdf5
        elif spath.endswith(".hdf5"):
            name_list, obj_list = []
            for k, v in obj.items():
                name_list.append(k)
                obj_list.append(v)
            with h5py.File(spath, "w") as hf:
                for (name, obj) in zip(name_list, obj_list):
                    hf.create_dataset(name, data=obj)
        # pth
        elif spath.endswith(".pth"):
            torch.save(obj, spath)

        # mat
        elif spath.endswith(".mat"):
            scipy.io.savemat(spath, obj)

        # catboost model
        elif spath.endswith(".cbm"):
            obj.save_model(spath)

        else:
            raise ValueError("obj was not saved.")

    except Exception as e:
        print(e)

    else:
        if show and not is_copying_files:
            # if show:
            print(f"\nSaved to: {spath}\n")


def check_encoding(file_path):
    from chardet.universaldetector import UniversalDetector

    detector = UniversalDetector()
    with open(file_path, mode="rb") as f:
        for binary in f:
            detector.feed(binary)
            if detector.done:
                break
    detector.close()
    enc = detector.result["encoding"]
    return enc


def is_listed_X(obj, types):
    """
    Example:
        obj = [3, 2, 1, 5]
        is_listed_X(obj,
    """
    import numpy as np

    try:
        conditions = []
        condition_list = isinstance(obj, list)

        if not (isinstance(types, list) or isinstance(types, tuple)):
            types = [types]

        _conditions_susp = []
        for typ in types:
            _conditions_susp.append(
                (np.array([isinstance(o, typ) for o in obj]) == True).all()
            )

        condition_susp = np.any(_conditions_susp)

        is_listed_X = np.all([condition_list, condition_susp])
        return is_listed_X

    except:
        return False


def save_listed_scalars_as_csv(
    listed_scalars,
    spath_csv,
    column_name="_",
    indi_suffix=None,
    round=3,
    overwrite=False,
    show=False,
):
    """Puts to df and save it as csv"""
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)
    indi_suffix = (
        np.arange(len(listed_scalars)) if indi_suffix is None else indi_suffix
    )
    df = pd.DataFrame(
        {"{}".format(column_name): listed_scalars}, index=indi_suffix
    ).round(round)
    df.to_csv(spath_csv)
    if show:
        print("\nSaved to: {}\n".format(spath_csv))


def save_listed_dfs_as_csv(
    listed_dfs,
    spath_csv,
    indi_suffix=None,
    overwrite=False,
    show=False,
):
    """listed_dfs:
        [df1, df2, df3, ..., dfN]. They will be written vertically in the order.

    spath_csv:
        /hoge/fuga/foo.csv

    indi_suffix:
        At the left top cell on the output csv file, '{}'.format(indi_suffix[i])
        will be added, where i is the index of the df.On the other hand,
        when indi_suffix=None is passed, only '{}'.format(i) will be added.
    """
    import numpy as np

    if overwrite == True:
        mv_to_tmp(spath_csv, L=2)

    indi_suffix = (
        np.arange(len(listed_dfs)) if indi_suffix is None else indi_suffix
    )
    for i, df in enumerate(listed_dfs):
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            i_suffix = indi_suffix[i]
            f_writer.writerow(["{}".format(indi_suffix[i])])
        df.to_csv(spath_csv, mode="a", index=True, header=True)
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            f_writer.writerow([""])
    if show:
        print("Saved to: {}".format(spath_csv))


def mk_mp4(fig, spath_mp4):
    from matplotlib import animation

    axes = fig.get_axes()

    def init():
        return (fig,)

    def animate(i):
        for ax in axes:
            ax.view_init(elev=10.0, azim=i)
        return (fig,)

    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=360, interval=20, blit=True
    )

    writermp4 = animation.FFMpegWriter(
        fps=60, extra_args=["-vcodec", "libx264"]
    )
    anim.save(spath_mp4, writer=writermp4)
    print("\nSaving to: {}\n".format(spath_mp4))


def mv_to_tmp(fpath, L=2):
    import os
    from shutil import move

    try:
        tgt_fname = connect_strs_with_hyphens(fpath.split("/")[-L:])
        tgt_fpath = "/tmp/{}".format(tgt_fname)
        move(fpath, tgt_fpath)
        print("Moved to: {}".format(tgt_fpath))
    except:
        pass


def save_optuna_study_as_csv_and_pngs(study, sdir):
    import optuna

    ## Trials DataFrame
    trials_df = study.trials_dataframe()

    ## Figures
    hparams_keys = list(study.best_params.keys())
    slice_plot = optuna.visualization.plot_slice(study, params=hparams_keys)
    contour_plot = optuna.visualization.plot_contour(
        study, params=hparams_keys
    )
    optim_hist_plot = optuna.visualization.plot_optimization_history(study)
    parallel_coord_plot = optuna.visualization.plot_parallel_coordinate(
        study, params=hparams_keys
    )
    hparam_importances_plot = optuna.visualization.plot_param_importances(
        study
    )
    figs_dict = dict(
        slice_plot=slice_plot,
        contour_plot=contour_plot,
        optim_hist_plot=optim_hist_plot,
        parallel_coord_plot=parallel_coord_plot,
        hparam_importances_plot=hparam_importances_plot,
    )

    ## Saves
    save(trials_df, sdir + "trials_df.csv")

    for figname, fig in figs_dict.items():
        save(fig, sdir + f"{figname}.png")
